#include <gtk/gtk.h>
#include <netdb.h>
#include <time.h>
#include <linux/ip.h>
#include <sys/ioctl.h>
#include <net/if.h>
#include <string.h>
#include <stdio.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "connection.h"
#include "interface.h"
#include "support.h"
#include "daemon.h"

GSList *conn_list = NULL;
guint conn_count = 0;
conn_info_t *conn_current = NULL;

static GtkWidget* get_window(void) {
    static GtkWidget *cw = NULL;

    if (!cw)
        cw = create_connection_window();

    return cw;
}

static gchar* icmp_type_str(guint8 t) {
    switch (t) {
        case 0: return "Echo reply";
        case 3: return "Destination unreachable";
        case 4: return "Source quench";
        case 5: return "Redirect";
        case 8: return "Echo request";
        case 9: return "Router advertisement";
        case 10: return "Router solicitation";
        case 13: return "Timestamp request";
        case 15: return "Information request";
        case 17: return "Address mask request";
        case 30: return "Traceroute";
        case 37: return "Domain name request";
        default: return "Unknown";
    }
}

static gchar* format_ip_addr(guint32 a) {
    static gchar txt[256];

    a = htonl(a);
    

    snprintf(txt, sizeof(txt), "%u.%u.%u.%u", a & 0xFF, (a >> 8) & 0xFF, (a >> 16) & 0xFF, a >> 24);
    return txt;
}


static gchar* format_host(guint32 a) {
    static gchar txt[256];
    struct hostent *he;

    a = htonl(a);
    
    if ((he = gethostbyaddr(&a, 4, AF_INET))) {
        g_strlcpy(txt, he->h_name, sizeof(txt));
        return txt;
    } else
        return format_ip_addr(ntohl(a));    
}


void conn_fill_info(conn_info_t *c) {
    struct protoent *pe = NULL;
    struct servent *se = NULL;
    static gchar txt[256];
    char *t;

    c->from_string = g_strdup(format_host(c->src_ip_address));
    c->to_string = g_strdup(format_host(c->dst_ip_address));

    if (c->proto == IPPROTO_ICMP)
        snprintf(t = txt, sizeof(txt), "ICMP %s", icmp_type_str(c->icmp_type));
    else if (c->proto == IPPROTO_TCP || c->proto == IPPROTO_UDP) {
        gchar *pp = c->proto == IPPROTO_TCP ? "TCP" : "UDP";
        
        if ((se = getservbyport(htons(c->port), NULL)))
            snprintf(t = txt, sizeof(txt), "%s Port#%u (%s)", pp, c->port, se->s_name);
        else
            snprintf(t = txt, sizeof(txt), "%s Port#%u", pp, c->port);
    } else if ((pe = getprotobynumber(c->proto)))
        t = pe->p_name;
    else
        t = "Unknown";

    c->type_string = g_strdup(t);

    strftime(txt, sizeof(txt), "%x %X", gmtime(&c->timestamp));
    c->timestamp_string = g_strdup(txt);
}

void conn_free(conn_info_t*c) {
    g_free(c->from_string);
    g_free(c->to_string);
    g_free(c->type_string);
    g_free(c);
}

void conn_show_info(conn_info_t *c) {
    GtkWidget* window;
    GtkTooltips *tooltips;
    static gchar txt[256];
    char *d;

    conn_fill_info(c);
    
    window = get_window();

    snprintf(txt, sizeof(txt), "<big><b>%s Connection</b></big>", c->direction == DIR_INCOMING ? "Incoming" : (c->direction == DIR_OUTGOING ? "Outgoing" : "Passing"));
    gtk_label_set_label(GTK_LABEL(lookup_widget(window, "title_label")), txt);

    gtk_image_set_from_stock(GTK_IMAGE(lookup_widget(window, "icon_image")), c->direction == DIR_INCOMING ? GTK_STOCK_GO_FORWARD : (c->direction == DIR_OUTGOING ? GTK_STOCK_GO_BACK : GTK_STOCK_NEW), GTK_ICON_SIZE_DIALOG);

    if (c->direction == DIR_INCOMING)
        d = c->in_device;
    else if (c->direction == DIR_OUTGOING)
        d = c->out_device;
    else
        snprintf(d = txt, sizeof(txt), "%s => %s", c->in_device, c->out_device);
    
    gtk_label_set_label(GTK_LABEL(lookup_widget(window, "interface_label")), d);
    gtk_label_set_label(GTK_LABEL(lookup_widget(window, "from_label")), c->from_string);
    gtk_label_set_label(GTK_LABEL(lookup_widget(window, "to_label")), c->to_string);
    gtk_label_set_label(GTK_LABEL(lookup_widget(window, "type_label")), c->type_string);

    tooltips = GTK_TOOLTIPS(lookup_widget(window, "tooltips"));
    gtk_tooltips_set_tip(tooltips, lookup_widget(window, "from_eventbox"), format_ip_addr(c->src_ip_address), NULL);
    gtk_tooltips_set_tip(tooltips, lookup_widget(window, "to_eventbox"), format_ip_addr(c->dst_ip_address), NULL);
    gtk_tooltips_set_delay(tooltips, 0);
    
    snprintf(txt, sizeof(txt), "%lu", c->id);
    gtk_label_set_label(GTK_LABEL(lookup_widget(window, "id_label")), txt);

    gtk_label_set_label(GTK_LABEL(lookup_widget(window, "time_label")), c->timestamp_string);

    gtk_misc_set_alignment(GTK_MISC(lookup_widget(get_window(), "count_label")), 0.5, 0.5);
    gtk_label_set_justify(GTK_LABEL(lookup_widget(get_window(), "count_label")), GTK_JUSTIFY_CENTER);

    gtk_widget_set_sensitive(lookup_widget(window, "reject_button"), !c->broadcast);

    gtk_label_set_label(GTK_LABEL(lookup_widget(window, "broadcast_label")), c->broadcast ? "Yes" : "No");

    gtk_widget_show_all(window);
    gtk_window_stick(GTK_WINDOW(window));
    gtk_window_set_position(GTK_WINDOW(window), GTK_WIN_POS_CENTER);
    gtk_window_present(GTK_WINDOW(window));
}

void conn_show_count() {
    static gchar txt[256];

    if (conn_count)
        snprintf(txt, sizeof(txt), "<b>Q:</b> %i", conn_count);
    else
        txt[0] = 0;

    gtk_label_set_label(GTK_LABEL(lookup_widget(get_window(), "count_label")), txt);
}

void conn_pop() {
    g_assert(!conn_current && conn_list);

    conn_current = (conn_info_t*) conn_list->data;
    conn_list = g_slist_remove(conn_list, conn_current);
    conn_count--;
    conn_show_count();
    conn_show_info(conn_current);
}

gboolean is_bc_addr(guint32 ip) {
    int sock, n;
    struct ifreq ifr[10];
    struct ifconf ifc;
    gboolean b = FALSE;
    
    if ((sock = socket(AF_INET, SOCK_DGRAM, 0)) < 0)
        return FALSE;

    ifc.ifc_req = ifr;
    ifc.ifc_len = sizeof(ifr);

    if (ioctl(sock, SIOCGIFCONF, &ifc) < 0) {
        b = FALSE;
        goto finish;
    }

    
    for (n = 0 ; n < ifc.ifc_len/sizeof(struct ifreq); n++) {
        if (ioctl(sock, SIOCGIFBRDADDR, &ifc.ifc_req[n]) >= 0) {
            if (htonl(ip) == ((struct sockaddr_in*) &ifc.ifc_req[n].ifr_broadaddr)->sin_addr.s_addr) {
                b = TRUE;
                break;
            }
        }
    }

finish:
    close(sock);
    return b;
}

static conn_info_t* ipq2ci(ipq_packet_msg_t* m) {
    conn_info_t* ci = g_new0(conn_info_t, 1);
    struct iphdr *ip = (struct iphdr*) m->payload;

    ci->id = m->packet_id;
    ci->direction = (m->indev_name[0] && m->outdev_name[0]) ? DIR_PASSING : (m->indev_name[0] ? DIR_INCOMING : DIR_OUTGOING);
    strncpy(ci->in_device, m->indev_name, IFNAMSIZ); ci->in_device[IFNAMSIZ] = 0;
    strncpy(ci->out_device, m->outdev_name, IFNAMSIZ); ci->out_device[IFNAMSIZ] = 0;

    if (m->data_len < sizeof(struct iphdr)) {
        g_message("Packet too small for complete IP header");
        g_free(ci);
        return NULL;
    }
    
    ci->src_ip_address = ntohl(ip->saddr);
    ci->dst_ip_address = ntohl(ip->daddr);
    ci->proto = ip->protocol;

    ci->broadcast = is_bc_addr(ci->dst_ip_address);

    ci->port = 0;
    ci->icmp_type = 0;
    
    if (ci->proto == IPPROTO_TCP || ci->proto == IPPROTO_UDP)
        ci->port = ntohs(*((guint16*) (m->payload + sizeof(struct iphdr) + 2)));
    else if (ci->proto == IPPROTO_ICMP)
        ci->icmp_type = *((guint8*) (m->payload + sizeof(struct iphdr)));
    
    ci->timestamp = m->timestamp_sec;
    
    return ci;
}

void conn_new(ipq_packet_msg_t *m) {
    conn_info_t *ci;

    if ((ci = ipq2ci(m))) {
        conn_list = g_slist_append(conn_list, ci);
        conn_count++;

        conn_show_count();
    
        if (!conn_current)
            conn_pop();
    }
}


void conn_verdict(verdict_t v) {
    guint r = 0;

    if (!conn_current)
        return;
    
    switch (v) {
        case VERDICT_REJECT : r = 13+2; break;
        case VERDICT_DROP : r = 0; break;
        case VERDICT_ACCEPT : r = 1; break;
    }

    daemon_verdict(conn_current->id, r);

    conn_free(conn_current);
    conn_current = NULL;
    
    if (!conn_list)
        gtk_widget_hide(get_window());
    else
        conn_pop();
}



